# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['selenium_testing_library']

package_data = \
{'': ['*']}

install_requires = \
['selenium>3.0.0', 'typing_extensions>=4.0.0']

setup_kwargs = {
    'name': 'selenium-testing-library',
    'version': '2022.13',
    'description': 'A Python Selenium library inspired by the Testing Library',
    'long_description': '# Selenium Testing Library\n\n[![PyPI version](https://badge.fury.io/py/selenium-testing-library.svg)](https://badge.fury.io/py/selenium-testing-library)\n[![test](https://github.com/anze3db/selenium-testing-library/actions/workflows/main.yml/badge.svg)](https://github.com/anze3db/selenium-testing-library/actions/workflows/main.yml) [![codecov](https://codecov.io/gh/anze3db/selenium-testing-library/branch/main/graph/badge.svg?token=L1M7HO3DL7)](https://codecov.io/gh/anze3db/selenium-testing-library)\n\nSelenium Testing Library (STL) is a Python library implementing [Testing-Library](https://testing-library.com/) in Selenium.\n\n## Dependencies\n\n- Python 3.7, 3.8, 3.9, 3.10, 3.11\n- [`selenium`](https://pypi.org/project/selenium/) >= 3.0.0, [`typing-extensions`](https://pypi.org/project/typing-extensions/) >= 4.0.0\n## Installation\n\n```\npip install selenium-testing-library\n```\n\n## Quick Start\n\n```python\nfrom selenium import webdriver\nfrom selenium_testing_library import Screen\n\ndriver = webdriver.Chrome()\ndriver.open(\'https://google.com/\')\n\nscreen = Screen(driver)\nsearch_input = screen.find_by_title("Search")\nsearch.send_keys("Dogs")\nsearch_button = screen.find_by_text("Google Search")\nsearch_button.click()\nscreen.wait_for_stale(search_button)\n```\n\n## Finding elements\n\nSTL implements the [Queries API](https://testing-library.com/docs/queries/about) from the Testing Library. The Testing Library queries `get_by`, `query_by`, `find_by`, and the multiple element equivalents `get_all_by`, `query_all_by`, `find_all_by` are used in places where you would normally use Selenium\'s `find_element` and `find_elements` functions.\n\n The difference between the different queries (`get_by`, `query_by`, `find_by`) is whether the query will throw an error if the element was not found (`get_by`), return `None` (`query_by`) or block, wait and retry until the element is found (`find_by`).\n\n * `get_by` returns the element matched and throws an exception if zero or more than one element matches. This is the main function that we should be using to locate elements on a page.\n * `query_by` returns the element matched or `None` if no element match. It throws an exception if more than one element matches. Mostly used for asserting that an element is **not** present: `assert not screen.query_by_text("not on page")`.\n * `find_by` behaves like `get_by`, but uses a `WebDriverWait` to wait until the element is present in the DOM.\n * `get_all_by` returns a list of elements matched. It raises an exception if no elements match.\n * `query_all_by` returns a list of elements matched. It returns an empty list when no elements match.\n * `find_all_by` behaves like `get_all_by`, but uses a `WebDriverWait` to wait until the elements are present in the DOM.\n\n When an element is found the queries return a Selenium [`WebElement`](https://selenium-python.readthedocs.io/api.html#module-selenium.webdriver.remote.webelement) or a list containing Selenium [WebElement](https://selenium-python.readthedocs.io/api.html#module-selenium.webdriver.remote.webelement)s when using `get_all_by`, `query_all_by`, `find_all_by`.\n\nThe queries accept a tuple containing the [By class identifier](https://selenium-python.readthedocs.io/api.html#locate-elements-by) and the search query, so they can be used with XPath, Css or any other native Selenium selector:\n\n```python\nfrom selenium import webdriver\nfrom selenium.webdriver.common.by import By\nfrom selenium_testing_library import Screen\n\nscreen = Screen(webdriver.Chrome())\nscreen.get_by((By.CSS, ".my_class")) # Will throw an exception if the element is not found\nscreen.query_by((By.ID, "my_id")) # you can use regular tuples as if you were using Selenium\'s find_element()\nscreen.find_by((By.XPATH, "//div"), timeout=5, poll_frequency=0.5) # locators for searching through text also work\n```\n\n## Locator Classes\n\nFor convenience Locator classes can be used instead of the tuples:\n\n```python\nfrom selenium import webdriver\nfrom selenium_testing_library import Screen, locators\n\nscreen = Screen(webdriver.Chrome())\nscreen.get_by(locators.Css(".my_class")) # Will throw an exception if the element is not found\nscreen.query_by(locators.Id("my_id")) # you can use regular tuples as if you were using Selenium\'s find_element()\nscreen.find_by(locators.XPath("//div"), timeout=5, poll_frequency=0.5) # locators for searching through text also work\n```\n\n## Testing Library Selectors\n\nBesides all the Selenium native By selectors, the queries also support Testing Library\'s selectors:\n * [Role](https://testing-library.com/docs/queries/byrole)\n * [LabelText](https://testing-library.com/docs/queries/bylabeltext)\n * [PlaceholderText](https://testing-library.com/docs/queries/byplaceholdertext)\n * [Text](https://testing-library.com/docs/queries/bytext)\n * [DisplayValue](https://testing-library.com/docs/queries/bydisplayvalue)\n * [AltText](https://testing-library.com/docs/queries/byalttext)\n * [Title](https://testing-library.com/docs/queries/bytitle)\n * [TestId](https://testing-library.com/docs/queries/bytestid)\n\n```python\nfrom selenium import webdriver\nfrom selenium_testing_library import Screen, locators\n\nscreen = Screen(webdriver.Chrome())\nscreen.get_by(locators.Text("My Text"))\nscreen.query_by(locators.Role("button", pressed=True))\nscreen.find_by(locators.TestId("my-test"), timeout=5, poll_frequency=0.5) # locators for searching through text also work\n```\n## Helper functions\n\nFor convenience helper functions on the screen class are available to avoid instantiating locator classes all over the place:\n\n[`screen.get_by_role(role_name)`](https://testing-library.com/docs/queries/byrole) Queries for elements with the given role.\n[`screen.get_by_label_text(text)`](https://testing-library.com/docs/queries/bylabeltext) Queries for label elements that match the text string and return the corresponding input element.\n[`screen.get_by_placeholder_text(text)`](https://testing-library.com/docs/queries/byplaceholdertext) Queries elements with the matching placeholder attribute.\n[`screen.get_by_text(text)`](https://testing-library.com/docs/queries/bytext) Queries elements where the content matches the provided text.\n[`screen.get_by_display_value(value)`](https://testing-library.com/docs/queries/bydisplayvalue) Queries inputs, textareas, or selects with matching display value.\n[`screen.get_by_alt_text(text)`](https://testing-library.com/docs/queries/byalttext) Queries elements with the matching alt attribute.\n`screen.get_by_title(text)` Queries elements with the matching title attribute.\n`screen.get_by_test_id(value)` Queries elements matching the `data-testid` value.\n`screen.get_by_css(css)` Queries elements matching the specified css selector.\n`screen.get_by_xpath(xpath)` Queries elements matching the specified xpath selector.\n\nThere are also `query_by_*`, `find_by_*`, `get_all_by_*`, `query_all_by_*`, `find_all_by_*`  equivalents.\n\n**Note:** The selenium project has removed the `find_element_by_*` and `find_elements_by_*` helper functions in the [Selenium 4.3.0](https://github.com/SeleniumHQ/selenium/releases/tag/selenium-4.3.0) release, so I just want to state that the `screen` helper functions will never be deprecated or removed.\n\nExamples:\n\n```python\nfrom selenium import webdriver\nfrom selenium_testing_library import Screen\n\nscreen = Screen(webdriver.Chrome())\nscreen.query_by_role("role_name")\nscreen.get_by_label_text("label text")\nscreen.find_all_by_text("my text", timeout=5, poll_frequency=0.5)\nscreen.get_all_by_alt_text("alt text")\n```\n\n## Wait functions\n\n`wait_for(condition_function)` Waits until the condition function returns a truthy value.\n`wait_for_stale(element)` Waits until the element is removed from the DOM.\n\nExamples:\n\n```python\nfrom selenium import webdriver\nfrom selenium_testing_library import Screen, locators\n\nscreen = Screen(webdriver.Chrome())\n\n# Wait for the element to be clickable:\nelement = screen.get_by_text("Submit")\nscreen.wait_for(lambda _: element.is_enabled(), timeout=5, poll_frequency=0.5)\n# Wait for the element to be removed from the page:\nscreen.wait_for_stale(element)\n```\n\n## Querying within elements\n\n`Within(element)` Used to limit the query to the children of the provided element\n\nExample:\n\n```python\nfrom selenium import webdriver\nfrom selenium_testing_library import Screen, Within\n\nscreen = Screen(webdriver.Chrome())\nparent_element = screen.get_by_css(".container")\nWithin(parent_element).get_by_title("My title inside the container")\n```\n\n# Contributing\n\nSetting up a local development environment\n\n```shell\ngit clone https://github.com/anze3db/selenium-testing-library.git && cd selenium-testing-library\npoetry install && poetry shell\n# Make sure `chromedriver` is in your PATH, download from https://chromedriver.chromium.org/downloads\n# run tests:\npytest --selenium-headless\n# run tests and display coverage info:\npytest --selenium-headless --cov=selenium_testing_library --cov-report html\n\n# To test on multiple Python versions make sure that py37, py38, py39 are\n# installed on your system and available through python3.7, python3.8,\n# python3.9. (Use pyenv and add the pyenv shims to your path\n# `export PATH=$(pyenv root)/shims:$PATH`). Then run tox:\ntox\n```\n\n# Releasing a new version\n\n```shell\nnpm run deploy\nbumpver update  # Wait and see if the CI is green\npoetry build && poetry publish\n```\n',
    'author': 'Anže Pečar',
    'author_email': 'anze@pecar.me',
    'maintainer': 'None',
    'maintainer_email': 'None',
    'url': 'https://github.com/anze3db/selenium-testing-library',
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'python_requires': '>=3.7,<4.0',
}


setup(**setup_kwargs)



import json
from pathlib import Path
from typing import Any, Callable, Generic, List, Optional, TypeVar, Union

from selenium.common.exceptions import (
    NoSuchElementException,
    TimeoutException,
    WebDriverException,
)
from selenium.webdriver.remote.webelement import WebElement
from selenium.webdriver.support import expected_conditions as EC
from selenium.webdriver.support.ui import WebDriverWait
from typing_extensions import Protocol

from . import locators

testing_library = (Path(__file__).parent / Path("main.js")).read_text()

Locator = locators.LocatorType

T = TypeVar("T")


class MultipleSuchElementsException(WebDriverException):
    ...


by_to_locator = {
    locators.By.CLASS_NAME: locators.ClassName,
    locators.By.CSS_SELECTOR: locators.Css,
    locators.By.ID: locators.Id,
    locators.By.LINK_TEXT: locators.LinkText,
    locators.By.PARTIAL_LINK_TEXT: locators.PartialLinkText,
    locators.By.TAG_NAME: locators.TagName,
    locators.By.XPATH: locators.XPath,
    locators.By.NAME: locators.Name,
    locators.By.TEXT: locators.Text,
    locators.By.ROLE: locators.Role,
    locators.By.PLACEHOLDER_TEXT: locators.PlaceholderText,
    locators.By.LABEL_TEXT: locators.LabelText,
    locators.By.ALT_TEXT: locators.AltText,
}


class ElementsFinder(Protocol):
    def find_elements(
        self, by: str = locators.By.ID, value: Optional[str] = None
    ) -> List[WebElement]:
        ...

    def execute_script(self, script: str, *args) -> List[WebElement]:
        ...


DriverType = TypeVar("DriverType", bound=ElementsFinder)


class Screen(Generic[DriverType]):
    def __init__(self, driver: DriverType):
        self.driver = driver
        self._finder: ElementsFinder = driver

    def _find_elements(self, locator: Locator) -> List[WebElement]:
        loc = self._ensure_locator(locator)
        if not isinstance(
            loc,
            (
                locators.Role,
                locators.Text,
                locators.PlaceholderText,
                locators.LabelText,
                locators.AltText,
                locators.Title,
                locators.TestId,
                locators.DisplayValue,
            ),
        ):
            return self._finder.find_elements(*loc)
        script_to_run = loc._testing_library_js_str()
        try:
            # Optimistically run the query, if __stl__ isn't defined on the page we'll get a JavaScript exception
            return self._finder.execute_script(script_to_run)
        except WebDriverException:
            # We assume that the error was `__stl__ is not defined` so we add __stl__ to the DOM and run the command again
            return self._finder.execute_script(f"{testing_library};{script_to_run}")

    def _ensure_locator(self, locator: Locator) -> locators.Locator:
        if isinstance(locator, locators.Locator):
            return locator
        by, selector = locator
        return by_to_locator[by](selector)

    def get_by(self, locator: Locator) -> WebElement:
        els = self._find_elements(locator)

        if not els:
            raise NoSuchElementException(self._get_no_element_message(locator))

        if len(els) > 1:
            raise MultipleSuchElementsException(
                self._get_multiple_elements_message(locator, els)
            )

        return els[0]

    def query_by(self, locator: Locator) -> Optional[WebElement]:
        els = self._find_elements(locator)

        if not els:
            return None
        if len(els) > 1:
            raise MultipleSuchElementsException(
                self._get_multiple_elements_message(locator, els)
            )

        return els[0]

    def find_by(
        self, locator: Locator, *, timeout: float = 5, poll_frequency: float = 0.5
    ) -> WebElement:
        try:
            els = self.wait_for(
                lambda _: self._find_elements(locator),
                timeout=timeout,
                poll_frequency=poll_frequency,
            )
        except TimeoutException:
            raise NoSuchElementException(self._get_no_element_message(locator))
        if len(els) > 1:
            raise MultipleSuchElementsException(
                self._get_multiple_elements_message(locator, els)
            )
        return els[0]

    def get_all_by(self, locator: Locator) -> List[WebElement]:
        els = self._find_elements(locator)
        if not els:
            raise NoSuchElementException(self._get_no_element_message(locator))

        return els

    def query_all_by(self, locator: Locator) -> List[WebElement]:
        try:
            return self.get_all_by(locator)
        except NoSuchElementException:
            return []

    def find_all_by(
        self, locator: Locator, *, timeout: float = 5, poll_frequency: float = 0.5
    ) -> List[WebElement]:
        try:
            return self.wait_for(
                lambda _: self._find_elements(locator),
                timeout=timeout,
                poll_frequency=poll_frequency,
            )
        except TimeoutException:
            raise NoSuchElementException(self._get_no_element_message(locator))

    ## Testing Library Selectors
    # By role
    def get_by_role(
        self,
        role: str,
        *,
        exact: bool = True,
        hidden: bool = False,
        name: Optional[str] = None,
        description: Optional[str] = None,
        selected: Optional[bool] = None,
        checked: Optional[bool] = None,
        pressed: Optional[bool] = None,
        current: Optional[Union[bool, str]] = None,
        expanded: Optional[bool] = None,
        queryFallbacks: Optional[bool] = None,
        level: Optional[int] = None,
    ) -> WebElement:
        return self.get_by(
            locators.Role(
                role,
                exact=exact,
                hidden=hidden,
                name=name,
                description=description,
                selected=selected,
                checked=checked,
                pressed=pressed,
                current=current,
                expanded=expanded,
                queryFallbacks=queryFallbacks,
                level=level,
            )
        )

    def query_by_role(
        self,
        role: str,
        *,
        exact: bool = True,
        hidden: bool = False,
        name: Optional[str] = None,
        description: Optional[str] = None,
        selected: Optional[bool] = None,
        checked: Optional[bool] = None,
        pressed: Optional[bool] = None,
        current: Optional[Union[bool, str]] = None,
        expanded: Optional[bool] = None,
        queryFallbacks: Optional[bool] = None,
        level: Optional[int] = None,
    ) -> Optional[WebElement]:
        return self.get_by(
            locators.Role(
                role,
                exact=exact,
                hidden=hidden,
                name=name,
                description=description,
                selected=selected,
                checked=checked,
                pressed=pressed,
                current=current,
                expanded=expanded,
                queryFallbacks=queryFallbacks,
                level=level,
            )
        )

    def find_by_role(
        self,
        role: str,
        *,
        exact: bool = True,
        hidden: bool = False,
        name: Optional[str] = None,
        description: Optional[str] = None,
        selected: Optional[bool] = None,
        checked: Optional[bool] = None,
        pressed: Optional[bool] = None,
        current: Optional[Union[bool, str]] = None,
        expanded: Optional[bool] = None,
        queryFallbacks: Optional[bool] = None,
        level: Optional[int] = None,
        timeout: float = 5,
        poll_frequency: float = 0.5,
    ) -> WebElement:
        return self.find_by(
            locators.Role(
                role,
                exact=exact,
                hidden=hidden,
                name=name,
                description=description,
                selected=selected,
                checked=checked,
                pressed=pressed,
                current=current,
                expanded=expanded,
                queryFallbacks=queryFallbacks,
                level=level,
            ),
            timeout=timeout,
            poll_frequency=poll_frequency,
        )

    def get_all_by_role(
        self,
        role: str,
        *,
        exact: bool = True,
        hidden: bool = False,
        name: Optional[str] = None,
        description: Optional[str] = None,
        selected: Optional[bool] = None,
        checked: Optional[bool] = None,
        pressed: Optional[bool] = None,
        current: Optional[Union[bool, str]] = None,
        expanded: Optional[bool] = None,
        queryFallbacks: Optional[bool] = None,
        level: Optional[int] = None,
    ) -> List[WebElement]:
        return self.get_all_by(
            locators.Role(
                role,
                exact=exact,
                hidden=hidden,
                name=name,
                description=description,
                selected=selected,
                checked=checked,
                pressed=pressed,
                current=current,
                expanded=expanded,
                queryFallbacks=queryFallbacks,
                level=level,
            )
        )

    def query_all_by_role(
        self,
        role: str,
        *,
        exact: bool = True,
        hidden: bool = False,
        name: Optional[str] = None,
        description: Optional[str] = None,
        selected: Optional[bool] = None,
        checked: Optional[bool] = None,
        pressed: Optional[bool] = None,
        current: Optional[Union[bool, str]] = None,
        expanded: Optional[bool] = None,
        queryFallbacks: Optional[bool] = None,
        level: Optional[int] = None,
    ) -> List[WebElement]:
        return self.query_all_by(
            locators.Role(
                role,
                exact=exact,
                hidden=hidden,
                name=name,
                description=description,
                selected=selected,
                checked=checked,
                pressed=pressed,
                current=current,
                expanded=expanded,
                queryFallbacks=queryFallbacks,
                level=level,
            )
        )

    def find_all_by_role(
        self,
        role: str,
        *,
        exact: bool = True,
        hidden: bool = False,
        name: Optional[str] = None,
        description: Optional[str] = None,
        selected: Optional[bool] = None,
        checked: Optional[bool] = None,
        pressed: Optional[bool] = None,
        current: Optional[Union[bool, str]] = None,
        expanded: Optional[bool] = None,
        queryFallbacks: Optional[bool] = None,
        level: Optional[int] = None,
        timeout: float = 5,
        poll_frequency: float = 0.5,
    ) -> List[WebElement]:
        return self.find_all_by(
            locators.Role(
                role,
                exact=exact,
                hidden=hidden,
                name=name,
                description=description,
                selected=selected,
                checked=checked,
                pressed=pressed,
                current=current,
                expanded=expanded,
                queryFallbacks=queryFallbacks,
                level=level,
            ),
            timeout=timeout,
            poll_frequency=poll_frequency,
        )

    # By text
    def get_by_text(
        self,
        text: str,
        *,
        selector: str = "*",
        exact: bool = True,
        ignore: Union[str, bool] = "script, style",
    ) -> WebElement:
        return self.get_by(
            locators.Text(text, selector=selector, exact=exact, ignore=ignore)
        )

    def query_by_text(
        self,
        text: str,
        *,
        selector: str = "*",
        exact: bool = True,
        ignore: Union[str, bool] = "script, style",
    ) -> Optional[WebElement]:
        return self.query_by(
            locators.Text(text, selector=selector, exact=exact, ignore=ignore)
        )

    def find_by_text(
        self,
        text: str,
        *,
        selector: str = "*",
        exact: bool = True,
        ignore: Union[str, bool] = "script, style",
        timeout: float = 5,
        poll_frequency: float = 0.5,
    ) -> WebElement:
        return self.find_by(
            locators.Text(text, selector=selector, exact=exact, ignore=ignore),
            timeout=timeout,
            poll_frequency=poll_frequency,
        )

    def get_all_by_text(
        self,
        text: str,
        *,
        selector: str = "*",
        exact: bool = True,
        ignore: Union[str, bool] = "script, style",
    ) -> List[WebElement]:
        return self.get_all_by(
            locators.Text(text, selector=selector, exact=exact, ignore=ignore)
        )

    def query_all_by_text(
        self,
        text: str,
        *,
        selector: str = "*",
        exact: bool = True,
        ignore: Union[str, bool] = "script, style",
    ) -> List[WebElement]:
        return self.query_all_by(
            locators.Text(text, selector=selector, exact=exact, ignore=ignore)
        )

    def find_all_by_text(
        self,
        text: str,
        *,
        selector: str = "*",
        exact: bool = True,
        ignore: Union[str, bool] = "script, style",
        timeout: float = 5,
        poll_frequency: float = 0.5,
    ) -> List[WebElement]:
        return self.find_all_by(
            locators.Text(text, selector=selector, exact=exact, ignore=ignore),
            timeout=timeout,
            poll_frequency=poll_frequency,
        )

    # By placeholder
    def get_by_placeholder_text(self, text: str, *, exact: bool = True) -> WebElement:
        return self.get_by(locators.PlaceholderText(text, exact=exact))

    def query_by_placeholder_text(
        self, text: str, *, exact: bool = True
    ) -> Optional[WebElement]:
        return self.query_by(locators.PlaceholderText(text, exact=exact))

    def find_by_placeholder_text(
        self,
        text: str,
        *,
        exact: bool = True,
        timeout: float = 5,
        poll_frequency: float = 0.5,
    ) -> WebElement:
        return self.find_by(
            locators.PlaceholderText(text, exact=exact),
            timeout=timeout,
            poll_frequency=poll_frequency,
        )

    def get_all_by_placeholder_text(
        self, text: str, *, exact: bool = True
    ) -> List[WebElement]:
        return self.get_all_by(locators.PlaceholderText(text, exact=exact))

    def query_all_by_placeholder_text(
        self, text: str, *, exact: bool = True
    ) -> List[WebElement]:
        return self.query_all_by(locators.PlaceholderText(text, exact=exact))

    def find_all_by_placeholder_text(
        self,
        text: str,
        *,
        exact: bool = True,
        timeout: float = 5,
        poll_frequency: float = 0.5,
    ) -> List[WebElement]:
        return self.find_all_by(
            locators.PlaceholderText(text, exact=exact),
            timeout=timeout,
            poll_frequency=poll_frequency,
        )

    # By label text
    def get_by_label_text(
        self, text: str, *, selector: str = "*", exact: bool = True
    ) -> WebElement:
        return self.get_by(locators.LabelText(text, selector=selector, exact=exact))

    def query_by_label_text(
        self, text: str, *, selector: str = "*", exact: bool = True
    ) -> Optional[WebElement]:
        return self.query_by(locators.LabelText(text, selector=selector, exact=exact))

    def find_by_label_text(
        self,
        text: str,
        *,
        selector: str = "*",
        exact: bool = True,
        timeout: float = 5,
        poll_frequency: float = 0.5,
    ) -> WebElement:
        return self.find_by(
            locators.LabelText(text, selector=selector, exact=exact),
            timeout=timeout,
            poll_frequency=poll_frequency,
        )

    def get_all_by_label_text(
        self, text: str, *, selector: str = "*", exact: bool = True
    ) -> List[WebElement]:
        return self.get_all_by(locators.LabelText(text, selector=selector, exact=exact))

    def query_all_by_label_text(
        self, text: str, *, selector: str = "*", exact: bool = True
    ) -> List[WebElement]:
        return self.query_all_by(
            locators.LabelText(text, selector=selector, exact=exact)
        )

    def find_all_by_label_text(
        self,
        text: str,
        *,
        selector: str = "*",
        exact: bool = True,
        timeout: float = 5,
        poll_frequency: float = 0.5,
    ) -> List[WebElement]:
        return self.find_all_by(
            locators.LabelText(text, selector=selector, exact=exact),
            timeout=timeout,
            poll_frequency=poll_frequency,
        )

    # By alt text
    def get_by_alt_text(self, text: str, *, exact: bool = True) -> WebElement:
        return self.get_by(locators.AltText(text, exact=exact))

    def query_by_alt_text(
        self, text: str, *, exact: bool = True
    ) -> Optional[WebElement]:
        return self.query_by(locators.AltText(text, exact=exact))

    def find_by_alt_text(
        self,
        text: str,
        *,
        exact: bool = True,
        timeout: float = 5,
        poll_frequency: float = 0.5,
    ) -> WebElement:
        return self.find_by(
            locators.AltText(text, exact=exact),
            timeout=timeout,
            poll_frequency=poll_frequency,
        )

    def get_all_by_alt_text(self, text: str, *, exact: bool = True) -> List[WebElement]:
        return self.get_all_by(locators.AltText(text, exact=exact))

    def query_all_by_alt_text(
        self, text: str, *, exact: bool = True
    ) -> List[WebElement]:
        return self.query_all_by(locators.AltText(text, exact=exact))

    def find_all_by_alt_text(
        self,
        text: str,
        *,
        exact: bool = True,
        timeout: float = 5,
        poll_frequency: float = 0.5,
    ) -> List[WebElement]:
        return self.find_all_by(
            locators.AltText(text, exact=exact),
            timeout=timeout,
            poll_frequency=poll_frequency,
        )

    # By title
    def get_by_title(self, title: str, *, exact: bool = True) -> WebElement:
        return self.get_by(locators.Title(title, exact=exact))

    def query_by_title(self, title: str, *, exact: bool = True) -> Optional[WebElement]:
        return self.query_by(locators.Title(title, exact=exact))

    def find_by_title(
        self,
        title: str,
        *,
        exact: bool = True,
        timeout: float = 5,
        poll_frequency: float = 0.5,
    ) -> WebElement:
        return self.find_by(
            locators.Title(title, exact=exact),
            timeout=timeout,
            poll_frequency=poll_frequency,
        )

    def get_all_by_title(self, title: str, *, exact: bool = True) -> List[WebElement]:
        return self.get_all_by(locators.Title(title, exact=exact))

    def query_all_by_title(self, title: str, *, exact: bool = True) -> List[WebElement]:
        return self.query_all_by(locators.Title(title, exact=exact))

    def find_all_by_title(
        self,
        title: str,
        *,
        exact: bool = True,
        timeout: float = 5,
        poll_frequency: float = 0.5,
    ) -> List[WebElement]:
        return self.find_all_by(
            locators.Title(title, exact=exact),
            timeout=timeout,
            poll_frequency=poll_frequency,
        )

    # By test id
    def get_by_test_id(self, text: str, *, exact: bool = True) -> WebElement:
        return self.get_by(locators.TestId(text, exact=exact))

    def query_by_test_id(
        self, text: str, *, exact: bool = True
    ) -> Optional[WebElement]:
        return self.query_by(locators.TestId(text, exact=exact))

    def find_by_test_id(
        self,
        text: str,
        *,
        exact: bool = True,
        timeout: float = 5,
        poll_frequency: float = 0.5,
    ) -> WebElement:
        return self.find_by(
            locators.TestId(text, exact=exact),
            timeout=timeout,
            poll_frequency=poll_frequency,
        )

    def get_all_by_test_id(self, text: str, *, exact: bool = True) -> List[WebElement]:
        return self.get_all_by(locators.TestId(text, exact=exact))

    def query_all_by_test_id(
        self, text: str, *, exact: bool = True
    ) -> List[WebElement]:
        return self.query_all_by(locators.TestId(text, exact=exact))

    def find_all_by_test_id(
        self,
        text: str,
        *,
        exact: bool = True,
        timeout: float = 5,
        poll_frequency: float = 0.5,
    ) -> List[WebElement]:
        return self.find_all_by(
            locators.TestId(text, exact=exact),
            timeout=timeout,
            poll_frequency=poll_frequency,
        )

    # By display value
    def get_by_display_value(self, value: str, *, exact: bool = True) -> WebElement:
        return self.get_by(locators.DisplayValue(value, exact=exact))

    def query_by_display_value(
        self, value: str, *, exact: bool = True
    ) -> Optional[WebElement]:
        return self.query_by(locators.DisplayValue(value, exact=exact))

    def find_by_display_value(
        self,
        value: str,
        *,
        exact: bool = True,
        timeout: float = 5,
        poll_frequency: float = 0.5,
    ) -> WebElement:
        return self.find_by(
            locators.DisplayValue(value, exact=exact),
            timeout=timeout,
            poll_frequency=poll_frequency,
        )

    def get_all_by_display_value(
        self, value: str, *, exact: bool = True
    ) -> List[WebElement]:
        return self.get_all_by(locators.DisplayValue(value, exact=exact))

    def query_all_by_display_value(
        self, value: str, *, exact: bool = True
    ) -> List[WebElement]:
        return self.query_all_by(locators.DisplayValue(value, exact=exact))

    def find_all_by_display_value(
        self,
        value: str,
        *,
        exact: bool = True,
        timeout: float = 5,
        poll_frequency: float = 0.5,
    ) -> List[WebElement]:
        return self.find_all_by(
            locators.DisplayValue(value, exact=exact),
            timeout=timeout,
            poll_frequency=poll_frequency,
        )

    ## Selenium Selectors
    # By css
    def get_by_css(self, css: str) -> WebElement:
        return self.get_by(locators.Css(css))

    def query_by_css(self, css: str) -> Optional[WebElement]:
        return self.query_by(locators.Css(css))

    def find_by_css(
        self, css: str, *, timeout: float = 5, poll_frequency: float = 0.5
    ) -> WebElement:
        return self.find_by(
            locators.Css(css), timeout=timeout, poll_frequency=poll_frequency
        )

    def get_all_by_css(self, css: str) -> List[WebElement]:
        return self.get_all_by(locators.Css(css))

    def query_all_by_css(self, css: str) -> List[WebElement]:
        return self.query_all_by(locators.Css(css))

    def find_all_by_css(
        self, css: str, *, timeout: float = 5, poll_frequency: float = 0.5
    ) -> List[WebElement]:
        return self.find_all_by(
            locators.Css(css), timeout=timeout, poll_frequency=poll_frequency
        )

    # By xpath
    def get_by_xpath(self, xpath: str) -> WebElement:
        return self.get_by(locators.XPath(xpath))

    def query_by_xpath(self, xpath: str) -> Optional[WebElement]:
        return self.query_by(locators.XPath(xpath))

    def find_by_xpath(
        self, xpath: str, *, timeout: float = 5, poll_frequency: float = 0.5
    ) -> WebElement:
        return self.find_by(
            locators.XPath(xpath), timeout=timeout, poll_frequency=poll_frequency
        )

    def get_all_by_xpath(self, xpath: str) -> List[WebElement]:
        return self.get_all_by(locators.XPath(xpath))

    def query_all_by_xpath(self, xpath: str) -> List[WebElement]:
        return self.query_all_by(locators.XPath(xpath))

    def find_all_by_xpath(
        self, xpath: str, *, timeout: float = 5, poll_frequency: float = 0.5
    ) -> List[WebElement]:
        return self.find_all_by(
            locators.XPath(xpath), timeout=timeout, poll_frequency=poll_frequency
        )

    # By id
    def get_by_id(self, id: str) -> WebElement:
        return self.get_by(locators.Id(id))

    def query_by_id(self, id: str) -> Optional[WebElement]:
        return self.query_by(locators.Id(id))

    def find_by_id(
        self, id: str, *, timeout: float = 5, poll_frequency: float = 0.5
    ) -> WebElement:
        return self.find_by(
            locators.Id(id), timeout=timeout, poll_frequency=poll_frequency
        )

    def get_all_by_id(self, id: str) -> List[WebElement]:
        return self.get_all_by(locators.Id(id))

    def query_all_by_id(self, id: str) -> List[WebElement]:
        return self.query_all_by(locators.Id(id))

    def find_all_by_id(
        self, id: str, *, timeout: float = 5, poll_frequency: float = 0.5
    ) -> List[WebElement]:
        return self.find_all_by(
            locators.Id(id), timeout=timeout, poll_frequency=poll_frequency
        )

    # By name
    def get_by_name(self, name: str) -> WebElement:
        return self.get_by(locators.Name(name))

    def query_by_name(self, name: str) -> Optional[WebElement]:
        return self.query_by(locators.Name(name))

    def find_by_name(
        self, name: str, *, timeout: float = 5, poll_frequency: float = 0.5
    ) -> WebElement:
        return self.find_by(
            locators.Name(name), timeout=timeout, poll_frequency=poll_frequency
        )

    def get_all_by_name(self, name: str) -> List[WebElement]:
        return self.get_all_by(locators.Name(name))

    def query_all_by_name(self, name: str) -> List[WebElement]:
        return self.query_all_by(locators.Name(name))

    def find_all_by_name(
        self, name: str, *, timeout: float = 5, poll_frequency: float = 0.5
    ) -> List[WebElement]:
        return self.find_all_by(
            locators.Name(name), timeout=timeout, poll_frequency=poll_frequency
        )

    # By tag name
    def get_by_tag_name(self, name: str) -> WebElement:
        return self.get_by(locators.TagName(name))

    def query_by_tag_name(self, name: str) -> Optional[WebElement]:
        return self.query_by(locators.TagName(name))

    def find_by_tag_name(
        self, name: str, timeout: float = 5, poll_frequency: float = 0.5
    ) -> WebElement:
        return self.find_by(
            locators.TagName(name), timeout=timeout, poll_frequency=poll_frequency
        )

    def get_all_by_tag_name(self, name: str) -> List[WebElement]:
        return self.get_all_by(locators.TagName(name))

    def query_all_by_tag_name(self, name: str) -> List[WebElement]:
        return self.query_all_by(locators.TagName(name))

    def find_all_by_tag_name(
        self, name: str, timeout: float = 5, poll_frequency: float = 0.5
    ) -> List[WebElement]:
        return self.find_all_by(
            locators.TagName(name), timeout=timeout, poll_frequency=poll_frequency
        )

    # By link text
    def get_by_link_text(self, value: str) -> WebElement:
        return self.get_by(locators.LinkText(value))

    def query_by_link_text(self, value: str) -> Optional[WebElement]:
        return self.query_by(locators.LinkText(value))

    def find_by_link_text(
        self, value: str, *, timeout: float = 5, poll_frequency: float = 0.5
    ) -> WebElement:
        return self.find_by(
            locators.LinkText(value), timeout=timeout, poll_frequency=poll_frequency
        )

    def get_all_by_link_text(self, value: str) -> List[WebElement]:
        return self.get_all_by(locators.LinkText(value))

    def query_all_by_link_text(self, value: str) -> List[WebElement]:
        return self.query_all_by(locators.LinkText(value))

    def find_all_by_link_text(
        self, value: str, *, timeout: float = 5, poll_frequency: float = 0.5
    ) -> List[WebElement]:
        return self.find_all_by(
            locators.LinkText(value), timeout=timeout, poll_frequency=poll_frequency
        )

    # By partial text
    def get_by_partial_link_text(self, text: str) -> WebElement:
        return self.get_by(locators.PartialLinkText(text))

    def query_by_partial_link_text(self, text: str) -> Optional[WebElement]:
        return self.query_by(locators.PartialLinkText(text))

    def find_by_partial_link_text(
        self, text: str, *, timeout: float = 5, poll_frequency: float = 0.5
    ) -> WebElement:
        return self.find_by(
            locators.PartialLinkText(text),
            timeout=timeout,
            poll_frequency=poll_frequency,
        )

    def get_all_by_partial_link_text(self, text: str) -> List[WebElement]:
        return self.get_all_by(locators.PartialLinkText(text))

    def query_all_by_partial_link_text(self, text: str) -> List[WebElement]:
        return self.query_all_by(locators.PartialLinkText(text))

    def find_all_by_partial_link_text(
        self, text: str, *, timeout: float = 5, poll_frequency: float = 0.5
    ) -> List[WebElement]:
        return self.find_all_by(
            locators.PartialLinkText(text),
            timeout=timeout,
            poll_frequency=poll_frequency,
        )

    # By class name
    def get_by_class_name(self, name: str) -> WebElement:
        return self.get_by(locators.ClassName(name))

    def query_by_class_name(self, name: str) -> Optional[WebElement]:
        return self.query_by(locators.ClassName(name))

    def find_by_class_name(
        self, name: str, *, timeout: float = 5, poll_frequency: float = 0.5
    ) -> WebElement:
        return self.find_by(
            locators.ClassName(name),
            timeout=timeout,
            poll_frequency=poll_frequency,
        )

    def get_all_by_class_name(self, name: str) -> List[WebElement]:
        return self.get_all_by(locators.ClassName(name))

    def query_all_by_class_name(self, name: str) -> List[WebElement]:
        return self.query_all_by(locators.ClassName(name))

    def find_all_by_class_name(
        self, name: str, *, timeout: float = 5, poll_frequency: float = 0.5
    ) -> List[WebElement]:
        return self.find_all_by(
            locators.ClassName(name),
            timeout=timeout,
            poll_frequency=poll_frequency,
        )

    def wait_for(
        self,
        method: Callable[[DriverType], T],
        *,
        timeout: float = 5,
        poll_frequency: float = 0.5,
        ignored_exceptions: Optional[bool] = None,
    ) -> T:
        return WebDriverWait(
            self.driver,
            timeout=timeout,
            poll_frequency=poll_frequency,
            ignored_exceptions=ignored_exceptions,
        ).until(method)

    def wait_for_stale(
        self, element: WebElement, *, timeout: float = 5, poll_frequency: float = 0.5
    ):
        return self.wait_for(
            EC.staleness_of(element), timeout=timeout, poll_frequency=poll_frequency
        )

    def _get_no_element_message(self, locator: Locator):
        return f"No element found with locator {locator}"

    def _get_multiple_elements_message(self, locator: Locator, els: List[WebElement]):
        el_str = ""
        for i, el in enumerate(els):
            el_str += f"{i}. {' '.join(el.get_attribute('outerHTML').splitlines())}\n"
        raise MultipleSuchElementsException(
            f"{len(els)} elements found with locator {locator}:\n{el_str}"
        )


class Within(Screen[WebElement]):
    def __init__(self, element: WebElement):
        self.element = element
        self._finder: ElementsFinder = element.parent

    def _find_elements(self, locator: Locator) -> List[WebElement]:
        loc = self._ensure_locator(locator)
        ex = "true" if loc.exact else "false"
        escaped_selector = json.dumps(loc.selector)

        if not isinstance(
            loc,
            (
                locators.Role,
                locators.Text,
                locators.PlaceholderText,
                locators.LabelText,
                locators.AltText,
                locators.Title,
                locators.TestId,
                locators.DisplayValue,
            ),
        ):
            return self.element.find_elements(*loc)

        script_to_run = loc._testing_library_js_str().replace(
            "(document", "(arguments[0]"
        )
        try:
            return self.element.parent.execute_script(
                script_to_run,
                self.element,
            )
        except WebDriverException:
            return self.element.parent.execute_script(
                f"{testing_library};{script_to_run}",
                self.element,
            )

    def wait_for(
        self,
        method: Callable[[WebElement], T],
        *,
        timeout: float = 5,
        poll_frequency: float = 0.5,
        ignored_exceptions: Optional[Any] = None,
    ) -> T:
        return WebDriverWait(
            self.element,
            timeout=timeout,
            poll_frequency=poll_frequency,
            ignored_exceptions=ignored_exceptions,
        ).until(method)

    def _get_no_element_message(self, locator: Locator):
        return f"No element found with locator {locator}:\n{self.element.get_attribute('outerHTML')}"


__all__ = [
    "Screen",
    "Within",
    "MultipleSuchElementsException",
    "NoSuchElementException",
    "locators",
]



import json
from typing import TYPE_CHECKING, Iterable, Optional, Union

from selenium.webdriver.common.by import By as SeleniumBy
from selenium.webdriver.remote.webelement import WebElement

if TYPE_CHECKING:
    from .screen import ElementsFinder


class By:
    CLASS_NAME: str = SeleniumBy.CLASS_NAME
    CSS_SELECTOR: str = SeleniumBy.CSS_SELECTOR
    ID: str = SeleniumBy.ID
    LINK_TEXT: str = SeleniumBy.LINK_TEXT
    PARTIAL_LINK_TEXT: str = SeleniumBy.PARTIAL_LINK_TEXT
    TAG_NAME: str = SeleniumBy.TAG_NAME
    XPATH: str = SeleniumBy.XPATH
    NAME: str = SeleniumBy.NAME
    # STL queries:
    ROLE = "role"
    TEXT = "text"
    PLACEHOLDER_TEXT = "placeholder text"
    LABEL_TEXT = "label text"
    ALT_TEXT = "alt text"
    TITLE = "title"
    TEST_ID = "test id"
    DISPLAY_VALUE = "display value"


class Locator:
    BY: str

    def __init__(self, selector: str, *, exact: bool = True):
        self.selector = selector
        self.exact = exact

    def __iter__(self):
        yield self.BY
        yield self.selector

    def __repr__(self):
        return f"{self.__class__.__name__}('{self.selector}', exact={self.exact})"


class Css(Locator):
    BY = By.CSS_SELECTOR


class XPath(Locator):
    BY = By.XPATH


class Id(Locator):
    BY = By.ID


class Name(Locator):
    BY = By.NAME


class TagName(Locator):
    BY = By.TAG_NAME


class LinkText(Locator):
    BY = By.LINK_TEXT


class PartialLinkText(Locator):
    BY = By.PARTIAL_LINK_TEXT


class ClassName(Locator):
    BY = By.CLASS_NAME


class Role(Locator):
    BY = By.ROLE

    def __init__(
        self,
        role: str,
        *,
        exact: bool = True,
        hidden: bool = False,
        name: Optional[str] = None,
        description: Optional[str] = None,
        selected: Optional[bool] = None,
        checked: Optional[bool] = None,
        pressed: Optional[bool] = None,
        current: Optional[Union[bool, str]] = None,
        expanded: Optional[bool] = None,
        queryFallbacks: Optional[bool] = None,
        level: Optional[int] = None,
    ):
        self.role = role
        self.exact = exact
        self.hidden = hidden
        self.name = name
        self.description = description
        self.selected = selected
        self.checked = checked
        self.pressed = pressed
        self.current = current
        self.expanded = expanded
        self.queryFallbacks = queryFallbacks
        self.level = level

    def __repr__(self):
        return f"""{self.__class__.__name__}(
    '{self.role}',
    exact={self.exact},
    hidden={self.hidden},
    name={self.name},
    description={self.description},
    selected={self.selected},
    checked={self.checked},
    pressed={self.pressed},
    current={self.current},
    expanded={self.expanded},
    queryFallbacks={self.queryFallbacks},
    level={self.level},
)"""

    def _testing_library_js_str(self):
        res = f"return __stl__.queryAllByRole(document, {json.dumps(self.role)}, {{ "
        res += f"exact: {json.dumps(self.exact)},"
        res += f"hidden: {json.dumps(self.hidden)},"
        if self.name is not None:
            res += f"name: {json.dumps(self.name)},"
        if self.description is not None:
            res += f"description: {json.dumps(self.description)},"
        if self.selected is not None:
            res += f"selected: {json.dumps(self.selected)},"
        if self.checked is not None:
            res += f"checked: {json.dumps(self.checked)},"
        if self.pressed is not None:
            res += f"pressed: {json.dumps(self.pressed)},"
        if self.current is not None:
            res += f"current: {json.dumps(self.current)},"
        if self.expanded is not None:
            res += f"expanded: {json.dumps(self.expanded)},"
        if self.queryFallbacks is not None:
            res += f"queryFallbacks: {json.dumps(self.queryFallbacks)},"
        if self.level is not None:
            res += f"level: {json.dumps(self.level)},"
        res += "});"
        return res


class Text(Locator):
    BY = By.TEXT

    def __init__(
        self,
        text: str,
        *,
        selector: str = "*",
        exact: bool = True,
        ignore: Union[str, bool] = "script, style",
    ):
        self.text = text
        self.selector = selector
        self.exact = exact
        self.ignore = ignore

    def __repr__(self):
        return f"""{self.__class__.__name__}(
    '{self.text}',
    selector='{self.selector}',
    exact={self.exact},
    ignore='{self.ignore}',
)"""

    def _testing_library_js_str(self):
        return f"return __stl__.queryAllByText(document, {json.dumps(self.text)}, {{selector: {json.dumps(self.selector)}, exact: {json.dumps(self.exact)}, ignore: {json.dumps(self.ignore)}}});"


class PlaceholderText(Locator):
    BY = By.PLACEHOLDER_TEXT

    def __init__(
        self,
        text: str,
        *,
        exact: bool = True,
    ):
        self.text = text
        self.exact = exact

    def __repr__(self):
        return f"{self.__class__.__name__}('{self.text}', exact={self.exact})"

    def _testing_library_js_str(self):
        return f"return __stl__.queryAllByPlaceholderText(document, {json.dumps(self.text)}, {{exact: {json.dumps(self.exact)}}});"


class LabelText(Locator):
    BY = By.LABEL_TEXT

    def __init__(
        self,
        text: str,
        *,
        selector: str = "*",
        exact: bool = True,
    ):
        self.text = text
        self.selector = selector
        self.exact = exact

    def __repr__(self):
        return f"{self.__class__.__name__}('{self.text}', selector={self.selector}, exact={self.exact})"

    def _testing_library_js_str(self):
        return f"return __stl__.queryAllByLabelText(document, {json.dumps(self.text)}, {{selector: {json.dumps(self.selector)}, exact: {json.dumps(self.exact)}}});"


class AltText(Locator):
    BY = By.ALT_TEXT

    def __init__(
        self,
        text: str,
        *,
        exact: bool = True,
    ):
        self.text = text
        self.exact = exact

    def __repr__(self):
        return f"{self.__class__.__name__}('{self.text}', exact={self.exact})"

    def _testing_library_js_str(self):
        return f"return __stl__.queryAllByAltText(document, {json.dumps(self.text)}, {{exact: {json.dumps(self.exact)}}});"


class Title(Locator):
    BY = By.TITLE

    def __init__(
        self,
        title: str,
        *,
        exact: bool = True,
    ):
        self.title = title
        self.exact = exact

    def __repr__(self):
        return f"{self.__class__.__name__}('{self.title}', exact={self.exact})"

    def _testing_library_js_str(self):
        return f"return __stl__.queryAllByTitle(document, {json.dumps(self.title)}, {{exact: {json.dumps(self.exact)}}});"


class TestId(Locator):
    BY = By.TITLE

    def __init__(
        self,
        text: str,
        *,
        exact: bool = True,
    ):
        self.text = text
        self.exact = exact

    def __repr__(self):
        return f"{self.__class__.__name__}('{self.text}', exact={self.exact})"

    def _testing_library_js_str(self):
        return f"return __stl__.queryAllByTestId(document, {json.dumps(self.text)}, {{exact: {json.dumps(self.exact)}}});"


class DisplayValue(Locator):
    BY = By.DISPLAY_VALUE

    def __init__(
        self,
        value: str,
        *,
        exact: bool = True,
    ):
        self.value = value
        self.exact = exact

    def __repr__(self):
        return f"{self.__class__.__name__}('{self.value}', exact={self.exact})"

    def _testing_library_js_str(self):
        return f"return __stl__.queryAllByDisplayValue(document, {json.dumps(self.value)}, {{exact: {json.dumps(self.exact)}}});"


LocatorType = Union[
    Iterable[
        str,
    ],
    Locator,
]
